import aiohttp
import aiohttp.web
import os
import importlib
import importlib.util

from apys import config, log, settings


class ApiObject:
    """
    Api properties that will be available on endpoints
    
    Properties:
        vars: dictionary api custom variables
        config: current config object
        supported_methods: supported http rest methods
        debug: function to log message
        error: function to log error
        (will be added endpoint required util modules)
    """

    def __init__(self, config_file, has_log=True):
        # Config file
        self.config_file = config_file
        self.config = config.load(config_file) if config_file else config.load()

        # Log
        self.bcolors = log.BColors if self.config['log']['color'] else log.NoColors

        # Supported methods
        self.supported_methods = [
            'get', 
            'post',
            'put',
            'delete',
            'head',
            'options',
            'default'
        ]

        # Shared variable
        self.vars = {}

        self.web = aiohttp.web

        ##
        # ADDING UTILS TO API PARAM
        #
        utils = {}
        if 'utils' in self.config:
            for util in self.config['utils']:
                if (
                        (os.path.isdir(os.path.join('.', settings.DIR_UTILS, util))) and
                        (os.path.exists(os.path.join('.', settings.DIR_UTILS, util, '__init__.py')))
                ):
                    if util not in utils:
                        spec = importlib.util.spec_from_file_location(
                            util,
                            os.path.join('.', settings.DIR_UTILS, util, '__init__.py'))

                        utils[util] = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(utils[util])

                        setattr(self, util, utils[util])

                        # calls util init function
                        if hasattr(utils[util], 'init'):
                            utils[util].init(self)

        if has_log:
            # Starts Logging Resources
            self.debug('')
            cors_url = self.config['server']['cors']
            if cors_url:
                self.debug('================== Resources ==================== {}cors-enabled=\'{}\'{}'.format(
                    self.bcolors.WARNING, cors_url, self.bcolors.ENDC))
            else:
                self.debug('================== Resources ====================')

            self.debug('')

            # Logging loaded utils
            for util in utils:
                self.debug('Util Loaded: [{}{}{}]'.format(self.bcolors.OKBLUE, util, self.bcolors.ENDC))

            self.debug('')

    def debug(self, msg, to='debug'):
        log.debug(self, msg, to)
        
    def error(self, msg, to='debug', ex=False):
        log.error(self, msg, to, ex)

    def __reduce__(self):
        return ApiObject, (self.config_file, False)



import argparse
import sys
import os
import importlib
import importlib.util

from apys import settings


def main():
    # Adding current dir to sys path
    sys.path.insert(0, './')

    parser = argparse.ArgumentParser(description='apys - a restful api framework.')
    parser.add_argument('-s', '--start', '--serve', 
                        action='store_true', default=False, 
                        help='starts server')
    parser.add_argument('-c', '--config',
                        action='store', default=False, 
                        help='select config file')
    parser.add_argument('-i', '--init',
                        action='store_true', default=False, 
                        help='initialize project folder')
    parser.add_argument('-b', '--build', 
                        action='store_true', default=False, 
                        help='build project [not yet implemented]')
    parser.add_argument('-v', '--version',
                        action='store_true', default=False, 
                        help='show lib version')

    ##
    # Adding CLI utils
    #
    utils = []
    if os.path.exists(os.path.join('.', settings.DIR_UTILS)):
        for subdir in os.listdir(os.path.join('.', settings.DIR_UTILS)):
            if(
                    (os.path.isdir(os.path.join('.', settings.DIR_UTILS, subdir))) and
                    (os.path.exists(os.path.join('.', settings.DIR_UTILS, subdir, '__cli__.py')))
            ):

                util = subdir
                spec = importlib.util.spec_from_file_location(
                    util,
                    os.path.join('.', settings.DIR_UTILS, util, '__cli__.py'))
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)

                # Get util CLI object
                cli = module.CLI(None)
                parser.add_argument('-{}'.format(util[0]), '--{}'.format(util),
                                    action=cli.action, default=cli.default,
                                    help=cli.help)

                # Adding util to list
                utils += [util]

    ##
    # Get arguments and execute scripts
    #
    args = parser.parse_args()

    if args.start:
        from apys import server
        server.start(args.config)
    elif args.init:
        from apys import project
        project.init()
    elif args.version:
        import apys
        print(apys.__version__)
    else:

        # start util script
        for util in utils:
            if getattr(args, util, False):

                # Cli argument
                arg = getattr(args, util, False)

                # Starting server
                from apys import server
                server.cli_start(util, args.config, arg)

                return

        # nothing found, show help
        parser.print_help()



import json
import os

from apys import settings

##
# CONSTANTS
#

_README = """# MY PROJECT

My project description

## LANGUAGE
[Python >= 3.4.2](https://docs.python.org/3/)

## LIBRARIES
* [apys](https://github.com/seijihirao/apys) - Backend restful framework

---

## RUNNING

```
$ apys -s
```
"""

##
# FUNCTIONS
#


def init():
    """
    Initiate project folders with sample config and readme
    """

    # folders
    dirs = [
        settings.DIR_CONFIG,
        settings.DIR_ENDPOINTS,
        settings.DIR_FILTERS,
        settings.DIR_UTILS
    ]
    __create_dirs(dirs)

    # configs
    __write_config_file(settings.DEFAULT_PROD_NAME, {
        'server': {
            'port': settings.DEFAULT_PROD_PORT,
            'cors': settings.DEFAULT_PROD_CORS
        },
        'log': {
            'file': {
                'debug': os.path.join(settings.DEFAULT_PROD_LOG_DIR, settings.DEFAULT_PROD_LOG_DEBUG_FILE),
                'error': os.path.join(settings.DEFAULT_PROD_LOG_DIR, settings.DEFAULT_PROD_LOG_ERROR_FILE)
            },
            'color': settings.DEFAULT_PROD_COLOR
        },
        'utils': []
    })

    __write_config_file(settings.DEFAULT_DEV_NAME, {
        'server': {
            'port': settings.DEFAULT_DEV_PORT,
            'cors': settings.DEFAULT_DEV_CORS
        },
        'utils': []
    })

    # readme
    with open(os.path.join('.', 'README.md'), 'w') as outfile:
        outfile.write(_README)


def __create_dirs(directories):
    for directory in directories:
        if not os.path.exists(os.path.join('.', directory)):
            os.makedirs(directory)


def __write_config_file(file, obj):
    with open(os.path.join('.', settings.DIR_CONFIG, '{}.json'.format(file)), 'w') as outfile:
        json.dump(obj, outfile, indent=4)



import os
import json

from apys import settings


def load(scope=None):
    """
    loads chosen config file
    if scope is not define loads default (first try `local`, then `dev` and last `prod`)
    
    Args:
        scope: config file, without `.json`

    Returns:
        obj: formatted json config object
    """
    if not scope:
        scope = default()

    path = os.path.join(settings.DIR_CONFIG, '{}.json'.format(scope))

    if not os.path.exists(path):
        raise EnvironmentError('No config file found for {}'.format(scope))

    with open(path, 'r') as config:
        obj = __get_env(json.loads(config.read()))
        obj['scope'] = scope

        # Default Values

        __fill_default_value(obj, 'log', {})
        __fill_default_value(obj['log'], 'file', settings.DEFAULT_LOG)
        if type(obj['log']['file']) == str:
            obj['log']['file'] = {
                'debug': obj['log']['file'],
                'error': obj['log']['file']
            }
        __fill_default_value(obj['log'], 'color', settings.DEFAULT_COLOR)

        __fill_default_value(obj, 'server', {})
        __fill_default_value(obj['server'], 'port', settings.DEFAULT_PORT)
        __fill_default_value(obj['server'], 'cors', settings.DEFAULT_CORS)

        return obj


def default():
    """
    Gets default configuration file

    Returns:
        scope: config filename
    """
    for scope in [settings.DEFAULT_LOCAL_NAME, settings.DEFAULT_DEV_NAME, settings.DEFAULT_PROD_NAME]:
        path = os.path.join(settings.DIR_CONFIG, '{}.json'.format(scope))
        if os.path.exists(path):
            return scope
    

def __fill_default_value(obj, key, default_value):
    """
    Fill Default value if key does not exists on object

    Args:
        obj: object to fill with default_value
        key: key that will be filled if nonexistent
        default_value: default_value value to be filled
    """
    if key not in obj:
        obj[key] = default_value


def __get_env(obj):
    """
    gets value from environment variable if `$` is preceded on value

    i.e.:

    ```
    {
        "key": "$API_KEY"
        "db_url": "$DB|localhost"
        "port": "$PORT|8080|int"
    }
    ```

     * `key` will now have the value of the `API_KEY` environment variable, or '$API_KEY' if env var does not exists
     * `db_url` will now have the value of the `DB` environment variable, or 'localhost' if env var does not exists
     * `port` will now have the value of the `PORT` environment variable, or 8080 if env var does not exists


    :param obj: object to format
    :return: formatted object
    """
    for key in obj:
        if type(obj[key]) == dict:
            obj[key] = __get_env(obj[key])
        elif type(obj[key]) == str and obj[key][0] == '$':
            env_var = obj[key][1:]

            # Case no default value
            if env_var.count('|') == 0:
                if env_var in os.environ:
                    obj[key] = os.environ[env_var]

            # Case default value
            if env_var.count('|') == 1:
                env_var, default_val = env_var.split('|')
                if env_var in os.environ:
                    obj[key] = os.environ[env_var]
                else:
                    obj[key] = default_val

            # Case default value with type
            if env_var.count('|') == 2:
                env_var, default_val, default_type = env_var.split('|')
                if env_var in os.environ:
                    obj[key] = os.environ[env_var]
                else:
                    obj[key] = eval(default_type)(default_val)
    return obj



import os
import importlib
import importlib.util
import re
import inspect

from collections import OrderedDict

from aiohttp import web
import aiohttp_cors

from apys import settings

#
# CONSTANTS
#


ATTR_FILTERS = settings.DIR_FILTERS
ATTR_UTILS = settings.DIR_UTILS

#
# Functions
#


def prepare(app, api, cors_url=''):
    """
    Mount routers and add them to aiohttp application 
    
    Returns:
        application (cyclone.web.Application): cyclone application with endpoint routers mounted 
    """

    if cors_url:
        cors = aiohttp_cors.setup(app, defaults={
            cors_url: aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                expose_headers='*',
                allow_headers='*'
            )
        })

    file_paths = []
    
    # read all endpoint files
    for root, subdirs, files in os.walk(os.path.join('.', settings.DIR_ENDPOINTS)):
        for file in files:
            if os.path.splitext(file)[1] == '.py':
                file_paths += [{
                    'url': os.path.splitext(os.path.relpath(os.path.join(root, file), settings.DIR_ENDPOINTS))[0],
                    'file': os.path.join(root, file)
                }]

    filters = {}

    # populate routes
    for file_path in file_paths:

        spec = importlib.util.spec_from_file_location(file_path['url'].replace('/', '-'), file_path['file'])
        file_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(file_module)

        supported_methods = api.supported_methods

        handler_props = {
            'api': api
        }

        loaded_methods = []
        loaded_filters = {}
        
        ##
        # ADDING FILTER TO LIST
        #

        def add_filters(cur_filter):
            if cur_filter not in filters:
                filter_spec = importlib.util.spec_from_file_location(
                    cur_filter,
                    '{}.py'.format(os.path.join(settings.DIR_FILTERS, *(cur_filter.split('.')))))

                filters[cur_filter] = importlib.util.module_from_spec(filter_spec)
                filter_spec.loader.exec_module(filters[cur_filter])

                # calls util init function
                if hasattr(filters[cur_filter], 'init'):
                    filters[cur_filter].init(api)

                loaded_filters[cur_filter] = filters[cur_filter]

        if hasattr(file_module, ATTR_FILTERS):
            for filt in file_module.filters:

                if type(filt) == str:
                    add_filters(filt)
                elif type(filt) == list:
                    for f in filt:
                        add_filters(f)


        ##
        # ADDING METHOD FUNCTIONS
        #
        def create_func(endpoint, cur_method):

            def exec_filter(req, cur_filter):
                # Calls 'always' filt function
                if hasattr(filters[cur_filter], 'always'):
                    filters[cur_filter].always(req, handler_props['api'])
                # Calls current method filt function
                if hasattr(filters[cur_filter], cur_method):
                    getattr(filters[cur_filter], cur_method)(req, handler_props['api'])

            ##
            # This is what gets executed when you call the endpoint
            #
            async def func(req):
                req.vars = {}
                try:
                    if req.has_body:
                        if req.content_type == 'application/json':
                            req.body = await req.json()
                        elif req.content_type == 'multipart/form-data' or \
                                req.content_type == 'application/x-www-form-urlencoded':
                            req.body = dict(await req.post())
                        else:
                            req.body = {}

                        req.body = __translate_json(req.body)
                    else:
                        req.body = {}
                except Exception as ex:
                    str_err = 'Error while getting json data'
                    api.error(str_err, ex=ex)
                    raise handler_props['api'].web.HTTPInternalServerError(reason=str_err)
                
                if hasattr(endpoint, ATTR_FILTERS):
                    for cur_filter in endpoint.filters:
                        if type(cur_filter) == str:
                            exec_filter(req, cur_filter)
                        elif type(cur_filter) == list:
                            exception = None
                            for f in cur_filter:
                                exception = None

                                try:
                                    exec_filter(req, f)
                                except Exception as ex:
                                    exception = ex

                                if exception is None:
                                    break

                            if exception is not None:
                                raise exception

                # Calls current method function
                
                method_function = getattr(endpoint, cur_method)
                if inspect.iscoroutinefunction(method_function):
                    return web.json_response(await method_function(req, handler_props['api']))
                else:
                    return web.json_response(method_function(req, handler_props['api']))

            return func

        # Adds route resource

        if cors_url:
            # noinspection PyUnboundLocalVariable
            resource = cors.add(app.router.add_resource('/{}'.format(file_path['url'])))
        else: 
            resource = app.router.add_resource('/{}'.format(file_path['url']))
        for method in supported_methods:
            if hasattr(file_module, method):

                loaded_methods += [method]  # Log purpose
                # Adds route
                if cors_url:
                    cors.add(resource.add_route(method.upper(), create_func(file_module, method)))
                else:
                    resource.add_route(method.upper(), create_func(file_module, method))
        
        # Logging loaded endpoints
        for method in loaded_methods:
            str_loaded_methods = '{}{}{}'.format(api.bcolors.OKGREEN, method, api.bcolors.ENDC)

            str_loaded_filters = ''
            for filt in loaded_filters:
                if hasattr(loaded_filters[filt], 'always'):
                    str_loaded_filters += '{}{}{}'.format(api.bcolors.OKBLUE, filt, api.bcolors.ENDC)
                    str_loaded_filters += ', '
                if hasattr(loaded_filters[filt], method):
                    str_loaded_filters += '{}{}{}'.format(api.bcolors.OKGREEN, filt, api.bcolors.ENDC)
                    str_loaded_filters += ', '
            str_loaded_filters = str_loaded_filters[0:-2]
            
            api.debug('Endpoint Loaded: [{}{}{}] ({}) <- {{{}}}'.format(
                api.bcolors.OKGREEN, file_path['url'], api.bcolors.ENDC, str_loaded_methods, str_loaded_filters))
    
    api.debug('')
    return file_paths


def __translate_json(obj):
    """
    Gets a dictionary and formats its keys,
    following the rules:

        {'var.0': 'value0', 'var.1': 'value1'}
        will be
        {'var': ['value0', 'value1']}

    And 
    
        {'var.key0': 'value0', 'var.key1': 'value1'}
        will be
        {'var': {'key0':'value0', 'key1':'value1'}}

    Args:
        obj - dictionary object to be converted
    
    Returns:
        translated object
    """
    # Only translates a dictionary
    if type(obj) != dict:
        return obj

    translated_obj = {}
    for key in obj:
        dot_count = key.count('.')
        if dot_count:
            keys = key.split('.')
            cur_translated_obj = translated_obj
            for cur_key in keys:
                if dot_count:
                    # Keeps iterating and adding nested objects
                    if cur_key not in cur_translated_obj:
                        cur_translated_obj[cur_key] = {}
                    cur_translated_obj = cur_translated_obj[cur_key]
                else:
                    # Last one should have value
                    cur_translated_obj[cur_key] = obj[key]
                dot_count -= 1
        else:
            translated_obj[key] = obj[key]

    return __dict_to_list(translated_obj)


def __dict_to_list(obj):
    """
    Converts a dictionary to a list recursively

        {'0':'value0', '1':'value1'}
        will be
        ['value0', 'value1']

    Args:
        obj - dictionary object to be converted

    Returns:
        list if conversion was possible
    """

    # Only converts a dictionary
    if type(obj) != dict or obj == {}:
        return obj

    # Recursively converts
    for key in obj:
        if type(obj[key]) == dict:
            obj[key] = __dict_to_list(obj[key])

    # Checks if it really can be converted to a list
    for key in obj:
        if not key.isdigit():
            return obj

    return list(OrderedDict(obj).values())  # needs an ordered dict to return list in order



import os

##
# CONFIG
#

DEFAULT_LOCAL_NAME = 'local'
DEFAULT_DEV_NAME = 'dev'
DEFAULT_PROD_NAME = 'prod'

DEFAULT_LOG = False
DEFAULT_COLOR = True
DEFAULT_PORT = '$PORT|8080|int'
DEFAULT_CORS = False

##
# PROJECT
#

DIR_CONFIG = 'config'
DIR_ENDPOINTS = 'endpoints'
DIR_FILTERS = 'filters'
DIR_UTILS = 'utils'

DEFAULT_PROD_PORT = '$PORT|80|int'
DEFAULT_PROD_CORS = False
DEFAULT_PROD_LOG_DIR = os.path.join('/', 'var', 'logs', 'apys')
DEFAULT_PROD_LOG_DEBUG_FILE = 'debug.log'
DEFAULT_PROD_LOG_ERROR_FILE = 'error.log'
DEFAULT_PROD_COLOR = DEFAULT_CORS

DEFAULT_DEV_PORT = DEFAULT_PORT
DEFAULT_DEV_CORS = '*'



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class BColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class NoColors:
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''
    BOLD = ''
    UNDERLINE = ''


def debug(api, message, to):
    """
    Publish a message to the global log publisher.
    
    Args:
        api: api object, implicitly passed
        message: the log message
        to: where to write message (defined on config)
    """
    if api.config['log']['file']:
        with open(api.config['log']['file'][to], 'a') as f:
            print(message, file=f)
    else:
        print(message)


def error(api, message, to, ex):
    """
    Publish an error to the global log publisher.
    
    Args:
        api: api object, implicitly passed
        message: the error message
        to: where to write message (defined on config)
        ex: the exception type (default=Exception)
    """
    if ex:
        message = '\n{}\n{}\n'.format(message, str(ex))

    message = '{}{}{}'.format(api.bcolors.FAIL, message, api.bcolors.ENDC)

    if api.config['log']['file']:
        with open(api.config['log']['file'][to], 'a') as f:
            print(message, file=f)
    else:
        print(message)



import sys
import os
import importlib
import importlib.util
import asyncio
import contextlib
from multiprocessing import Process, Event, Manager


from aiohttp import web

from apys import apiobject, routes, config, settings


def start(config_file=None, process={'event': False, 'dict': {}}):
    """
    Starts server

    Args:
        config_file: default to 'prod'
        disable_log: True to disable access logs
        process: multiprocessing object
            process['event']: Event object (used to wait)
            process['dict']: Manager.dict object
    """

    # Preparing environment
    app = web.Application()
    global_api = apiobject.ApiObject(config_file)
    process['dict']['api'] = global_api

    # Setting log files
    if global_api.config['log']['file']:
        f_out = open(global_api.config['log']['file']['debug'], 'a')
        sys.stdout = f_out

        f_err = open(global_api.config['log']['file']['error'], 'a')
        sys.stderr = f_err

    # Preparing routes
    process['dict']['endpoints'] = routes.prepare(app, global_api, cors_url=global_api.config['server']['cors'])

    # Start logging
    start_str = 'Using [' + global_api.bcolors.HEADER + '{}' + global_api.bcolors.ENDC + '] configuration\n'
    if config_file:
        global_api.debug(start_str.format(config_file))
    else:
        global_api.debug(start_str.format(config.default()))

    if process['event']:
        process['event'].set()

    # Running server
    web.run_app(app, port=global_api.config['server']['port'])


def cli_start(util_name, args_config=None, arg=None):
    """
    Starts cli util on a process and server on another

    :param util_name: util dir name
    :param args_config: config file name got from cli args
    :param arg: argument used on cli
    :return:
    """
    if os.path.exists(os.path.join('.', settings.DIR_UTILS)):
        if (
                (os.path.isdir(os.path.join('.', settings.DIR_UTILS, util_name))) and
                (os.path.exists(os.path.join('.', settings.DIR_UTILS, util_name, '__cli__.py')))
        ):
            spec = importlib.util.spec_from_file_location(
                util_name,
                os.path.join('.', settings.DIR_UTILS, util_name, '__cli__.py'))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            cli = module.CLI(arg)

            # Getting config file
            if args_config:
                config_file = args_config
            elif hasattr(cli, 'config'):
                config_file = cli.config
            else:
                config_file = None

            process = {
                'dict': Manager().dict(),
                'event': Event()
            }

            def start_server(server_config_file, multiprocess):
                with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):  # disable logs
                    start(server_config_file, multiprocess)

            server_process = Process(target=start_server, args=(config_file, process))
            server_process.start()
            process['event'].wait()

            # execute async function
            loop = asyncio.new_event_loop()
            loop.run_until_complete(cli.start(process['dict']['api'], process['dict']['endpoints']))

            server_process.terminate()



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

try:
    import apys
    version = apys.__version__
except (IOError, ImportError):
    version = 'Error, could not get version'

setup(
    name='apys',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='A simple back-end rest framework in python using aiohttp lib',
    long_description=long_description,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://github.com/seijihirao/apys',

    # Author details
    author='Rodrigo Seiji Piubeli Hirao',
    author_email='rodrigo.seiji.hirao@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='backend rest web service',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['demo']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['aiohttp', 'aiohttp_cors'],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'apys=apys.commandline:main'
        ],
    },
)

# Requirements to Build project
# # pandoc - $ sudo pacman -S pandoc
# # pyandoc - $ pip install pyandoc
#
# Build project
# # python setup.py sdist
#
# Create wheel
# # python setup.py bdist_wheel
#
# Upload to Pypi
# # twine upload dist/apys-x.y.z-py3-none-any.whl
# # # x.y.z - is the actual version